#include <process.h>
#include <sys/uio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/neutrino.h>
#include <spawn.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <Ph.h>
#include <Pt.h>
#include <errno.h>
#include <pthread.h>

#include "def.h"

int Flickerless = 0;
PtWidget_t *Window, *RealWindow;


int CreateWidget (tMsg &msg, int rcvid)
  {
  PtWidget_t *widget, *widparent;  
  PtWidgetClassRef_t *wc;
  static PtArg_t arg [5] = {
    { Pt_ARG_GRAPHIC_FLAGS, Pt_FLOAT_POS, Pt_TRUE },
    { Pt_ARG_COLOR, 0, 0 },
    { Pt_ARG_POINTS, 0, 0 }
    };
  static PtArg_t argc [] = {
    { Pt_ARG_POS, 0, 0 },
    { Pt_ARG_GROUP_ORIENTATION, Pt_GROUP_ASIS, 0 }
    };
  static PtArg_t argt [] = {
    { Pt_ARG_POS, 0, 0 },
    { Pt_ARG_COLOR, 0, 0 },
    { Pt_ARG_TEXT_STRING, 0, 0 }
    };
  static PtArg_t argpd [] = {
    { Pt_ARG_POS, 0, 0 },
    { Pt_ARG_DIM, 0, 0 }
    };
  static PtArg_t argi [] = {
    { Pt_ARG_LABEL_TYPE, Pt_IMAGE, 0 },
    { Pt_ARG_POS, 0, 0 },
    { Pt_ARG_LABEL_IMAGE, 0, 0}
    };

  PtArg_t *args = arg;
  int nargs = 3;
      
  if (msg.parent == 0) widparent = Window;
  else widparent = (PtWidget_t*)msg.parent;

  switch (msg.subtype)
    {
    case PIXEL:
      {
      tPixelMsg *m = (tPixelMsg*)&msg;
      wc = PtPixel;
      arg[1].value = m->color;
      arg[2].value = (int)&(m->p);
      arg[2].len = 1;
      break;
      }
    case LINE:
      {
      tLineMsg *m = (tLineMsg*)&msg;
      wc = PtLine;
      arg[1].value = m->color;
      arg[2].value = (int)&(m->p1);
      arg[2].len = 2;
      break;
      }
    case POLYLINE:
      {
      tPolyMsg *m = (tPolyMsg*)&msg;
      wc = PtPolygon;
      arg[1].value = m->color;
      arg[2].value = (int)(m->p);
      arg[2].len = m->pnum;
      break;
      }
    case POLYGON:
      {
      tPolyMsg *m = (tPolyMsg*)&msg;
      wc = PtPolygon;
      arg[1].value = m->color;
      arg[2].value = (int)(m->p);
      arg[2].len = m->pnum;
      arg[3].type = Pt_ARG_POLYGON_FLAGS;
      arg[3].value = Pg_CLOSED;
      arg[3].len = Pt_TRUE;
      nargs++;
      break;
      }
    case RECT:
      {
      tRectMsg *m = (tRectMsg*)&msg;
      wc = PtRect;
      arg[1].value = m->color;
      arg[2].value = (int)&(m->p1);
      arg[2].len = 2;
      arg[3].type = Pt_ARG_RECT_ROUNDNESS;
      arg[3].value = m->roundness;
      nargs++;
      break;
      }
    case ELLIPSE:
      {
      tLineMsg *m = (tLineMsg*)&msg;
      wc = PtEllipse;
      arg[1].value = m->color;
      arg[2].value = (int)&(m->p1);
      arg[2].len = 2;
      break;
      }
    case ARC:
      {
      tArcMsg *m = (tArcMsg*)&msg;
      static int conv [] = { Pt_ARC_CURVE, Pt_ARC_PIE, Pt_ARC_CHORD };
      wc = PtArc;
      arg[1].value = m->color;
      arg[2].value = (int)&(m->p1);
      arg[2].len = 2;
      arg[3].type = Pt_ARG_ARC_START;
      arg[3].value = m->sangle;
      arg[4].type = Pt_ARG_ARC_END;
      arg[4].value = m->eangle;
      arg[5].type = Pt_ARG_ARC_TYPE;
      arg[5].value = conv [m->arctype];
      nargs += 3;
      break;
      }
    case GRID:
      {
      tGridMsg *m = (tGridMsg*)&msg;
      wc = PtGrid;
      arg[1].value = m->color;
      arg[2].value = (int)&(m->p1);
      arg[2].len = 2;
      arg[3].type = Pt_ARG_GRID_HORIZONTAL;
      arg[3].value = m->nrows;
      arg[4].type = Pt_ARG_GRID_VERTICAL;
      arg[4].value = m->ncols;
      nargs += 2;
      break;
      }
    case TEXT:
      {
      tTextMsg *m = (tTextMsg*)&msg;
      wc = PtLabel;
      argt[0].value = (int)&(m->p);
      argt[1].value = m->color;
      argt[2].value = (int)m->str;
      args = argt;
      break;
      }
    case IMAGE32:
      {
      tImageMsg *m = (tImageMsg*)&msg;
      wc = PtLabel;
      argi[1].value = (int)&(m->p1);
      PhImage_t * im = PhCreateImage (0, m->p2.x, m->p2.y, Pg_IMAGE_DIRECT_8888, 0, 0, 1);
      im->bpl = m->p2.x * 4;
      MsgRead (rcvid, im->image, m->p2.x * m->p2.y * 4, sizeof(tImageMsg));
      argi[2].value = (int)im;
      args = argi;
      break;
      }
    case IMAGE24:
      {
      tImageMsg *m = (tImageMsg*)&msg;
      wc = PtLabel;
      argi[1].value = (int)&(m->p1);
      PhImage_t * im = PhCreateImage (0, m->p2.x, m->p2.y, Pg_IMAGE_DIRECT_888, 0, 0, 1);
      im->bpl = m->p2.x * 3;
      MsgRead (rcvid, im->image, m->p2.x * m->p2.y * 3, sizeof(tImageMsg));
      argi[2].value = (int)im;
      args = argi;
      break;
      }
    case IMAGE32SM:
      {
      tImageSMMsg *m = (tImageSMMsg*)&msg;
      tImageSM *p;
      wc = PtLabel;
      argi[1].value = (int)&(m->p);
      PhImage_t * im = PhCreateImage (0, m->w, m->h, Pg_IMAGE_DIRECT_8888, 0, 0, 1);
      im->bpl = m->w * 4;
      m->fd = shm_open (m->name, O_RDONLY, 0666);
      im->image = (char*) mmap (0, m->w * m->h * 4, PROT_READ, MAP_SHARED, m->fd, sizeof(tImageSM));
      argi[2].value = (int)im;
      args = argi;
      break;
      }
    case GROUP:
      {
      tPointMsg *m = (tPointMsg*)&msg;
      wc = PtGroup;
      argc[0].value = (int)&(m->p);
      args = argc;
      nargs = 2;
      break;
      }
    }
  PtEnter (Pt_EVENT_PROCESS_PREVENT);     
  widget = PtCreateWidget (wc, widparent, nargs, args);
  if (widget)     
    if ((msg.parent == 0) && (msg.subtype != GROUP)) 
      PtRealizeWidget (widget), PtFlush ();
  PtLeave (Pt_EVENT_PROCESS_PREVENT);   
  return (int)widget;    
  }


void ChangeWidget (tMsg &msg)
  {
  PtWidget_t *widget;  
  int nargs = 1;
  PtArg_t args[2];
  tPoint p [2];
    
  widget = (PtWidget_t*)msg.parent;
  if (widget == 0) widget = RealWindow;

  args[0].len = args[1].len = 0;

  switch (msg.subtype)
    {
    case COLOR:
      {
      tColorMsg *m = (tColorMsg*)&msg;
      args[0].type = Pt_ARG_COLOR;
      args[0].value = m->color;
      break;
      }
    case FILL:
      {
      tColorMsg *m = (tColorMsg*)&msg;
      if (m->parent) args[0].type = Pt_ARG_INSIDE_COLOR;
      else args[0].type = Pt_ARG_FILL_COLOR, widget = Window;
      args[0].value = m->color;
      break;
      }
    case MOVETO:
      {
      tPointMsg *m = (tPointMsg*)&msg;
      args[0].type = Pt_ARG_POS;
      args[0].value = (int)&(m->p);
      break;
      }
    case MOVE:
      {
      tPointMsg *m = (tPointMsg*)&msg;
      args[0].type = Pt_ARG_POS;
      args[0].value = 0;
      PtGetResources (widget, 1, args);
      p[0] = *((tPoint*)args[0].value);
      p[0].x += m->p.x;
      p[0].y += m->p.y;
      args[0].value = (int)p;
      break;
      }
    case ENLARGETO:
      {
      tPolyMsg *m = (tPolyMsg*)&msg;
      if (m->parent)
        {
        args[0].type = Pt_ARG_POINTS;
        args[0].value = (int)(m->p);
        args[0].len = m->pnum;
        }
      else // Window
        {
        args[0].type = Pt_ARG_POS;
        args[0].value = (int)&(m->p[0]);
        args[1].type = Pt_ARG_DIM;
        m->p[1].x -= m->p[0].x;
        m->p[1].y -= m->p[0].y;
        args[1].value = (int)&(m->p[1]);
        nargs++;
        }
      break;
      }
    case ENLARGE:
      {
      tLineMsg *m = (tLineMsg*)&msg;
      args[0].type = Pt_ARG_POINTS;
      args[0].value = 0;
      PtGetResources (widget, 1, args);
      p[0] = *((tPoint*)args[0].value);
      p[1] = *((tPoint*)args[0].value+1);
      p[0].x -= m->p1.x;
      p[0].y -= m->p1.y;
      p[1].x += m->p2.x;
      p[1].y += m->p2.y;
      args[0].value = (int)p;
      break;
      }
    case LINEWIDTH:
      {
      tColorMsg *m = (tColorMsg*)&msg;
      args[0].type = Pt_ARG_LINE_WIDTH;
      args[0].value = m->color;
      break;
      }
    case CHTEXT:
      {
      tTextMsg *m = (tTextMsg*)&msg;
      args[0].type = Pt_ARG_TEXT_STRING;
      args[0].value = (int)m->str;
      break;
      }
    }
  PtEnter (Pt_EVENT_PROCESS_PREVENT);     
  PtSetResources(widget, nargs, args);
  PtRealizeWidget (widget); 
//  PtFlush ();
  PtLeave (Pt_EVENT_PROCESS_PREVENT);   
  }


int GetResource (tMsg &msg)
  {
  PtWidget_t *widget;  
  int *res;
  int stat;
      
  widget = (PtWidget_t*)msg.parent;
  if (widget == 0) widget = RealWindow;

  switch (msg.subtype)
    {
    case GETCOLOR:
      {
      stat = PtGetResource (widget, Pt_ARG_COLOR, &res, 0);
      break;
      }
    case GETFILL:
      {
      if (msg.parent)
        stat = PtGetResource (widget, Pt_ARG_INSIDE_COLOR, &res, 0);
      else
        stat = PtGetResource (Window, Pt_ARG_FILL_COLOR, &res, 0);
      break;
      }
    case GETPOS:
      {
      stat = PtGetResource (widget, Pt_ARG_POS, &res, 0);
      break;
      }
    case GETDIM:
      {
      stat = PtGetResource (widget, Pt_ARG_DIM, &res, 0);
      break;
      }
    }
  if (stat == 0) return *res;
  else return -1;
  }

void ManageWidget (tMsg &msg)
  {
  PtWidget_t *widget;  
      
  widget = (PtWidget_t*)msg.parent;
  if (widget == 0) widget = Window;

  PtEnter (Pt_EVENT_PROCESS_PREVENT);        
  switch (msg.subtype)
    {
    case REALIZE:   PtRealizeWidget   (widget); break;
    case UNREALIZE: PtUnrealizeWidget (widget); break;
    case DELETE:    PtDestroyWidget   (widget); break;
    case CLEAR:     PtClearWidget     (widget); break;
    }
  PtFlush ();
  PtLeave (Pt_EVENT_PROCESS_PREVENT);      
  }


struct tIQE // Input Queue Element
  {
  tIQE *next;
  int rcvid;
  int pri;
  };
pthread_mutex_t IQL = PTHREAD_MUTEX_INITIALIZER;
tIQE InputQueue = {0, 0, 0};

void EnqueueInput (int rcvid)
  {
  tIQE *p, *pp, *q;
  _msg_info info; MsgInfo (rcvid, &info);
  q = new tIQE; q->rcvid = rcvid, q->pri = info.priority;
  pthread_mutex_lock (&IQL);
  p = InputQueue.next, pp = &InputQueue;
  while ((p) && (p->pri >= q->pri)) pp = p, p = p->next;
  q->next = p, pp->next = q;
  pthread_mutex_unlock (&IQL);
  }

int KeyPressed (PtWidget_t *win, void *foo, PtCallbackInfo_t *info)
  {
  PhKeyEvent_t *ev;
  tIQE *p;
  ev = (PhKeyEvent_t*)PhGetData (info->event);
  if (ev->key_flags & Pk_KF_Sym_Valid)
    {
    pthread_mutex_lock (&IQL);
    p = InputQueue.next; if (p) InputQueue.next = p->next;
    pthread_mutex_unlock (&IQL);
    if (p)
      {
      MsgReply (p->rcvid, 0, &(ev->key_sym), 1);
      delete p;
      }
    }
  return Pt_CONSUME;
  }

int DequeueInput (int rcvid)
  {
  tIQE *p, *pp;
  pthread_mutex_lock (&IQL);
  p = InputQueue.next, pp = &InputQueue;
  while ((p) && (p->rcvid != rcvid)) pp = p, p = p->next;
  if (p) pp->next = p->next;
  pthread_mutex_unlock (&IQL);
  if (p == 0) return -1;
  delete p;
  return 0;
  }

char file [32];

int fd;

void* LibvgServer (char *name)
  {
  union
    {
    char buf [100];
    _pulse pls;
    tMsg msg;
    };
  int pid, chid, rcvid;
  int ident;
      
  chid = ChannelCreate (_NTO_CHF_UNBLOCK |_NTO_CHF_DISCONNECT | _NTO_CHF_COID_DISCONNECT);
  pid = getpid ();
  strcpy (file, "/dev/shmem/vg-");
  if (name == 0) strcat (file, "VG");
  else strncat (file, name, 13);
  fd = open (file, O_CREAT+O_RDWR, 0666);
  write (fd, &pid, 4), write (fd, &chid, 4);
//  unlink (file);
  while(1)
    {
    rcvid = MsgReceive (chid, buf, sizeof(buf), 0);
    if (rcvid == -1)   break;
    if (rcvid == 0) // Pulse received
         {
//         printf ("vg: pulse received code = %d\n", pls.code);
         if (pls.code == _PULSE_CODE_DISCONNECT)
            ConnectDetach(pls.scoid);
         if (pls.code == _PULSE_CODE_UNBLOCK)
            {
            ident = DequeueInput (pls.value.sival_int);
            MsgError (pls.value.sival_int, 0);
            if (ident < 0) perror ("vg: strange unblock");
            }
         if (pls.code == 0)  unlink (file), PtExit (0);
         continue;
         }      

      ident = 0;
      if (msg.type == CREATE) ident = CreateWidget (msg, rcvid);
      else if (msg.type == CHANGE) ChangeWidget (msg);
      else if (msg.type == GETRES) ident = GetResource (msg);
      else if (msg.type == MANAGE) ManageWidget (msg);
      else if (msg.type == INPUT)
        {
        EnqueueInput (rcvid);
        continue;
        }
      else ident = -1;

      MsgReply (rcvid, 0, &ident, 4);
    }
  
  perror ("vg: MsgReceive");
  unlink (file);
  PtExit (EXIT_FAILURE);
  return 0;
  }

int WindowDestroyed (PtWidget_t *win, void *foo, PtCallbackInfo_t *info)
  {
  unlink (file);
  return Pt_CONTINUE;
  }

int WindowRezised (PtWidget_t *win, void *foo, PtCallbackInfo_t *info)
  {
  PtContainerCallback_t *q = (PtContainerCallback_t*)info->cbdata;
  PtSetResource (Window, Pt_ARG_DIM, &q->new_dim, 0);
//  printf ("resized: w=%d h=%d\n", q->new_dim.w, q->new_dim.h);
  return Pt_CONTINUE;
  }


int FindOpt (char c, char *argv[])
  {
  int i = 1;
  while (argv[i]) 
    if (argv[i][1] == c) return i;
    else i++;
  return 0;
  }


void* IamHere (void* foo)
  {
  int t;
  while (1)
    {
    t = time (0);
    pwrite (fd, &t, 4, 8);
    sleep (1);
    }
  }


int main (int argc, char *argv[])
  {
  PtArg_t args[5];
  PhArea_t area = {100, 100, 640, 480};
  int bc = 0; // PgRGB (10, 20, 30);  
  const char *name = "VG", *vgosc;
  int i;
  
  i = fork ();
  if (i > 0) return 0;

  vgosc = getenv ("VGOSC");
  if (vgosc)
    {
    if (vgosc[0] == '1') Flickerless = 1;
    else Flickerless = 0;
    }

  i = FindOpt ('a', argv);  
  if (i) // format: -a X Y W H
    area.pos.x = atoi (argv[i+1]),
    area.pos.y = atoi (argv[i+2]),
    area.size.w = atoi (argv[i+3]),
    area.size.h = atoi (argv[i+4]);
  i = FindOpt ('c', argv);
  if (i) // format: -c R G B
    bc = PgRGB (atoi(argv[i+1]), atoi(argv[i+2]), atoi(argv[i+3]));
  i = FindOpt ('n', argv);
  if (i) // format: -n NAME
    name = argv[i+1];
    
  PtSetArg (&args[0], Pt_ARG_FILL_COLOR, bc, 0);
  PtSetArg (&args[1], Pt_ARG_AREA, &area, 0);
  PtSetArg (&args[2], Pt_ARG_FLAGS, Pt_CALLBACKS_ACTIVE | Pt_GETS_FOCUS, Pt_TRUE);
  PtSetArg (&args[3], Pt_ARG_WINDOW_TITLE, name, 0);
  PtSetArg (&args[4], Pt_ARG_RESIZE_FLAGS, Pt_FALSE, Pt_RESIZE_XY_BITS);
  Window = RealWindow = PtAppInit (0, 0, 0, 5, args);
  if (RealWindow == 0) perror ("vg: AppInit"), PtExit (EXIT_FAILURE);
  
  PtAddFilterCallback (RealWindow, Ph_EV_KEY, KeyPressed, 0);

  PtAddCallback (RealWindow, Pt_CB_IS_DESTROYED, WindowDestroyed, 0);

  if (Flickerless)
    {
    PtAddCallback (RealWindow, Pt_CB_RESIZE, WindowRezised, 0);
    PtSetArg (&args[0], Pt_ARG_RESIZE_FLAGS, Pt_FALSE, Pt_RESIZE_XY_BITS);
    PtSetArg (&args[1], Pt_ARG_FILL_COLOR, bc, 0);
    Window = PtCreateWidget (PtOSContainer, RealWindow, 2, args);
    if (Window == 0) perror ("vg: PtOSContainer"), PtExit (EXIT_FAILURE);
    }
  
  PtRealizeWidget (RealWindow); 
  pthread_create (0, 0, (void*(*)(void*))LibvgServer, (void*)name);    
  pthread_create (0, 0, IamHere, 0);
  PtMainLoop();
  return 0;
  }
